{$M $1000,0,0}

uses reloc, dos, env;

type device_header_ptr = ^device_header;
     device_header =
       record
         next_driver:   device_header_ptr;
         attribute:     word;
         ofs_strategy:  word;
         ofs_interrupt: word;
         name:          array [1..8] of char;
       end;

     ptrrec     =  record
                     pofs,
                     pseg:  word;
                   end;

     device_request_header =
       record
       case Integer of
       0: (
         length:          byte;
         unit_code:       byte;    {block devices}
         command_code:    byte;
         status:          word;
         reserved:        array [1..8] of byte;
         nr_of_units:     byte;    {at install time, block devices}
         end_of_driver:   pointer; {at install time}
         config_line_ptr: pointer; {at install time, ends with CR}
       );
       1: (
         length_2:        byte;
         unit_code_2:     byte;
         command_code_2:  byte;
         status_2:        word;
         reserved_2:      array [1..8] of byte;
         nr_of_units_2:   byte;
         buffer_addres:   pointer; {pointer to time structure}
         number:          byte;    {length of time structure}
       );
       end;

     exe_header =  record
                     exe_signature,
                     length_im_m512,
                     file_size_512p,
                     nr_of_rel_items,
                     header_size,
                     min_nr_of_pars,
                     max_nr_of_pars,
                     offset_stack_seg,
                     initial_sp,
                     exe_checksum,
                     exe_entry_point,
                     offset_code_seg,
                     offset_first_rel_item,
                     overlay_nr:              word;
                   end;

const uninstall: boolean = false;
      quiet:     boolean = false;

var ptr_to_dos_clock_ptr,
    nul_device_header_ptr,
    our_device_header_ptr: device_header_ptr;
    request_header:        device_request_header;
    dos_segment:           word;
    intr_ptr,
    strat_ptr:             pointer;
    sys_command_line:      string;
    timer_counter:         longint absolute $40:$6C;

procedure check_for_compression;
var head:      exe_header;
    f:         file;
    br_result: word;
    image_st:  pointer;
begin
  our_device_header_ptr:= ptr(prefixseg + $10, 0);
  if our_device_header_ptr^.next_driver <> ptr($FFFF,$FFFF)
  then begin writeln('No device driver in front!'); halt($FF) end;
  filemode:= 0; {only read access}
  assign(f, getprogramname);
  {$I-}
  reset (f,1);
  {$I+}
  if ioresult <> 0
  then begin writeln('Error checking for compression'); halt($FE) end;
  blockread(f, head, sizeof(head), br_result);
  seek(f, (head.header_size shl 4));
  blockread(f, image_st, sizeof(image_st), br_result);
  close(f);
  if image_st <> ptr($FFFF,$FFFF)
  then
  begin
    writeln('Warning: program is compressed and should as such not be used in config.sys!');
  end;
end;

procedure do_preliminary_work;
begin
  check_for_compression;
  sys_command_line:='DEVICE=' +
    getprogramname + ' ' + string(ptr(prefixseg,$80)^) + #13#10#0;
  uninstall:= (pos('/u', string(ptr(prefixseg,$80)^)) <> 0)
              or (pos('/u', string(ptr(prefixseg,$80)^)) <> 0);
  quiet:= (pos('/q', string(ptr(prefixseg,$80)^)) <> 0)
          or (pos('/Q', string(ptr(prefixseg,$80)^)) <> 0);
  our_device_header_ptr:= ptr(prefixseg + $10, 0);
  strat_ptr:= ptr(ptrrec(our_device_header_ptr).pseg, our_device_header_ptr^.ofs_strategy);
  intr_ptr:=  ptr(ptrrec(our_device_header_ptr).pseg, our_device_header_ptr^.ofs_interrupt);
  longint(request_header.config_line_ptr):= succ(longint(@sys_command_line));
  request_header.command_code:= 1; {init}
  asm
    mov   ah, 52h
    int   21h
    mov   [dos_segment], es
    mov   [ptrrec(nul_device_header_ptr).pseg], es
    mov   [ptrrec(nul_device_header_ptr).pofs], bx
    mov   [ptrrec(ptr_to_dos_clock_ptr).pseg], es
    add   bx, 8
    mov   [ptrrec(ptr_to_dos_clock_ptr).pofs], bx
  end;
  if lo(ver) < 3 then inc(ptrrec(nul_device_header_ptr).pofs, $17) {2.x}
  else if (lo(ver) = 3) and (hi(ver) = 0)
  then inc(ptrrec(nul_device_header_ptr).pofs, $28) {3.0}
  else inc(ptrrec(nul_device_header_ptr).pofs, $22); {3.1+}
end;

procedure install_sys;
var pars_needed,
    memory_block:     word;
    counter_value:    longint;
    int_1A_pointer,
    int_8_pointer:    pointer;
begin
  asm
    mov   bx, seg @data
    mov   es, bx
    mov   bx, offset request_header
    call  [strat_ptr]
    call  [intr_ptr]
  end;
  pars_needed:= (ptrrec(request_header.end_of_driver).pofs + 15) shr 4;
  if pars_needed = 0 then halt(0);
  asm
    mov   [doserror], 0
    mov   ah, 48h
    mov   bx, [pars_needed]
    int   21h
    jc    @@error
    mov   [memory_block], ax
    jmp   @@end
  @@error:
    mov   [doserror], ax
  @@end:
  end;
  if doserror <> 0
  then begin writeln('Not enough memory'); halt($FD) end;
  make_mem_block_independent(ptr(memory_block, 0));
  getintvec($1A, int_1A_pointer);
  getintvec($8,  int_8_pointer);
  ptrrec(int_1A_pointer).pseg:= memory_block;
  ptrrec(int_8_pointer).pseg:= memory_block;
  counter_value:= timer_counter;
  while counter_value = timer_counter do
  begin end; {wait until a tick has happened}
  move(our_device_header_ptr^, ptr(memory_block, 0)^, pars_needed shl 4);
  setintvec($1A, int_1A_pointer);
  setintvec($8,  int_8_pointer);
  ptrrec(our_device_header_ptr).pseg:= memory_block;
  our_device_header_ptr^.next_driver:= nul_device_header_ptr^.next_driver;
  nul_device_header_ptr^.next_driver:= our_device_header_ptr;
  ptr_to_dos_clock_ptr^.next_driver:= our_device_header_ptr;
end;

procedure uninstall_sys;
var memory_block:        word;
    int_8_pointer:       pointer;
    t_device_header_ptr: device_header_ptr;
begin
  getintvec($8,  int_8_pointer);
  memory_block:= ptrrec(int_8_pointer).pseg;
  asm
    mov   bx, seg @data
    mov   es, bx
    mov   bx, offset request_header
    call  [strat_ptr]
    call  [intr_ptr]
  end;
  if request_header.status = $810C {general failure}
  then halt(2);
  ptrrec(our_device_header_ptr).pseg:= memory_block;
  {walk the chain}
  t_device_header_ptr:= nul_device_header_ptr;
  while (ptrrec(t_device_header_ptr^.next_driver).pofs <> $FFFF)
    and (t_device_header_ptr^.next_driver <> our_device_header_ptr)
  do begin t_device_header_ptr:= t_device_header_ptr^.next_driver end;
  {did we find our stuff?}
  if t_device_header_ptr^.next_driver <> our_device_header_ptr
  then begin writeln('Error uninstalling'); halt($F6) end;
  {unlink}
  t_device_header_ptr^.next_driver:= our_device_header_ptr^.next_driver;
  {walk the chain again}
  t_device_header_ptr:= nul_device_header_ptr;
  while (ptrrec(t_device_header_ptr^.next_driver).pofs <> $FFFF)
    and (t_device_header_ptr^.attribute <> $8008)
  do begin t_device_header_ptr:= t_device_header_ptr^.next_driver end;
  {did we find a clock device?}
  if t_device_header_ptr^.attribute <> $8008
  then begin writeln('Error uninstalling'); halt($F5) end;
  {let DOS point to this old clock driver}
  ptr_to_dos_clock_ptr^.next_driver:= t_device_header_ptr;
  asm     {release the block}
    mov   ah, 49h
    mov   es, [memory_block]
    int   21h
  end;
end;

begin
  do_preliminary_work;
  if uninstall
  then uninstall_sys
  else install_sys
end.