{Patches a device driver (or any other) file in front of an exe file}

{Works only if the 2 files apart are not bigger than 64 kB}

type exe_header =  record
                     exe_signature,
                     length_im_m512,
                     file_size_512p,
                     nr_of_rel_items,
                     header_size,
                     min_nr_of_pars,
                     max_nr_of_pars,
                     offset_stack_seg,
                     initial_sp,
                     exe_checksum,
                     exe_entry_point,
                     offset_code_seg,
                     offset_first_rel_item,
                     overlay_nr:              word;
                   end;

     ptrrec     =  record
                     pofs,
                     pseg:  word;
                   end;

var inexefilename,
    outexefilename,
    sysfilename:       string;
    inexefile,
    outexefile,
    sysfile:           file;
    infile_ptr,
    sysfile_ptr,
    outfile_ptr:       pointer;
    ehp:               ^exe_header;


procedure help;
begin
  writeln('Patchsys, patches a device driver file in front of an exe file');
  writeln;
  writeln('Usage:       Patchsys exefile sysfile [outfile]');
  writeln;
  writeln('    exefile  file to be patched with the sysfile');
  writeln('    sysfile  device driver file to be put in front');
  writeln('    outfile  result file (default is same as exefile)');
  writeln;
  writeln('Works only if the 2 files apart are not bigger than 64 kB');
  writeln;
  halt(0)
end;

procedure do_work;
var insize,
    syssize,
    or_image_size,
    outsize:         longint;
    extra_pars:      word;
    reloc_items:     word;
    reloc_table_ptr: pointer;
    r_addres:        pointer;
    t:               word;
begin
  filemode:= 0; {read only}
  {read exe file}
  assign(inexefile, inexefilename);
  {$I-}
  reset (inexefile, 1);
  {$I+}
  if ioresult <> 0
  then begin writeln(inexefilename, ' not found'); halt(1) end;
  insize:= filesize(inexefile);
  getmem(infile_ptr, insize);
  blockread(inexefile, infile_ptr^, insize);
  close(inexefile);
  if word(infile_ptr^) <> $5A4D {'MZ'}
  then begin writeln(inexefilename, ' is not an exe file'); halt(3) end;
  {read sys file}
  assign(sysfile, sysfilename);
  {$I-}
  reset (sysfile, 1);
  {$I+}
  if ioresult <> 0
  then begin writeln(sysfilename, ' not found'); halt(2) end;
  syssize:= filesize(sysfile);
  getmem(sysfile_ptr, syssize);
  blockread(sysfile, sysfile_ptr^, syssize);
  close(sysfile);
  syssize:= (syssize + 15) and $FFF0;  {round up to paragraphs}
  extra_pars:= syssize shr 4;
  outsize:= syssize + insize;
  ehp:= infile_ptr;
  with ehp^ do
  begin
    length_im_m512:= outsize mod 512;
    file_size_512p:= outsize div 512;
    if length_im_m512 <> 0 then inc(file_size_512p);
    inc(min_nr_of_pars, extra_pars);
    if max_nr_of_pars < ((640 * 1024)/16)
    then inc(max_nr_of_pars, extra_pars);
    inc(offset_stack_seg, extra_pars);
    inc(offset_code_seg, extra_pars);
    reloc_items:= nr_of_rel_items;
    reloc_table_ptr:= ehp;
    inc(longint(reloc_table_ptr), offset_first_rel_item);
    or_image_size:= insize - (header_size shl 4);
  end;
  for t:= 1 to reloc_items do
  begin
    r_addres:= pointer(reloc_table_ptr^);
    inc(ptrrec(r_addres).pseg, ptrrec(ehp).pseg);
    inc(ptrrec(r_addres).pofs, ptrrec(ehp).pofs);
    inc(ptrrec(r_addres).pseg, ehp^.header_size);
    inc(word(r_addres^), extra_pars);
    inc(ptrrec(reloc_table_ptr^).pseg, extra_pars);
    inc(longint(reloc_table_ptr), 4);
  end;
  filemode:= 1; {write only}
  assign(outexefile, outexefilename);
  rewrite(outexefile,1);
  blockwrite(outexefile, infile_ptr^, ehp^.header_size shl 4);
  blockwrite(outexefile, sysfile_ptr^, syssize);
  inc(ptrrec(infile_ptr).pseg, ehp^.header_size);
  blockwrite(outexefile, infile_ptr^, insize - (ehp^.header_size shl 4));
  close(outexefile);
end;

begin
  if paramcount = 2
  then
  begin
    inexefilename:= paramstr(1);
    sysfilename:= paramstr(2);
    outexefilename:= inexefilename;
  end
  else if paramcount = 3
  then
  begin
    inexefilename:= paramstr(1);
    sysfilename:= paramstr(2);
    outexefilename:= paramstr(3);
  end
  else help;
  do_work;
end.