;Filename	psp_bda.inc
;
;Description	This include file contains structures and segment
;		descriptions to get easy access	to the Program Segment
;		Prefix, the Bios Data Area and Memory Control Blocks.
;		A few equates and macros are also included.
;
;Macros
;
;TP_Str		Allocate a string using Turbo Pascal format
;call_far	Create storage space for a dword and generate code for making
;		a far call to it
;jmp_far	Create storage space for a dword and generate code for making
;		a far jump to it
;
;Structures
;
;far_ptr	gives easy access to a pointer's offset & segment
;curs_ptr	gives easy access to a cursor position's column & row
;pascal_string	representing a Turbo Pascal string
;
;Segments
;
;PSP		describing the structure of the Program Segment Prefix
;BDA		describing video related variables in the BIOS Data Area
;MCB		describing the structure of Memory Control Blocks


false		=	0
true		=	not false

CR		equ	13		;carriage return
LF		equ	10		;line feed
Newline		equ	CR,LF		;carriage return linefeed pair

far_ptr		struc			;gives easy access to a pointer's
o		dw	?		;offset &
s		dw	?		;segment
far_ptr		ends

pascal_string	struc			;Structure representing a TP string
length_byte	db	?		; Leading length byte(<= 255)
contents	db	255 dup (?)	; Contents of the string
pascal_string	ends

call_far	macro pointer		;create storage space for a dword and
		db	9Ah		; generate code for making a far call
pointer		dd	?		; to it
		endm

jmp_far		macro pointer		;create storage space for a dword and
		db	0EAh		; generate code for making a far jump
pointer		dd	?		; to it
		endm

curs_ptr	struc			;gives easy access to cursor
col		db	?		;column &
row		db	?		;row
curs_ptr	ends			;coordinate

bda		segment at 40h		;Bios Data Area,
		org	1Ah
kbd_head	dw	?		;pointer to head of keyboard buffer
kbd_tail	dw	?		;pointer to tail of keyboard buffer
		org	49h		; with some video related information
video_mode	db	?		;current video mode
columns		dw	?		;number columns on screen
bytes_p_screen	dw	?		;bytes per screen
screen_offset	dw	?		;offset of start of video buffer
cursor_loc	dw	?		;cursor location on video page 0
cursor_loc_1	dw	?		;cursor location on video page 1
cursor_loc_2	dw	?		;cursor location on video page 2
cursor_loc_3	dw	?		;cursor location on video page 3
cursor_loc_4	dw	?		;cursor location on video page 4
cursor_loc_5	dw	?		;cursor location on video page 5
cursor_loc_6	dw	?		;cursor location on video page 6
cursor_loc_7	dw	?		;cursor location on video page 7
cursor_mode	equ	this word	;end & start of cursor shape
cursor_end	db	?		;end of cursor (bottomline)
cursor_start	db	?		;start of cursor (upperline)
active_page	db	?		;active video page
addr_crtc	dw	?		;I/O-port address of CRTC
crt_mode_set	db	?		;6845 CRT mode control register value
crt_palette	db	?		;value of CGA color select register
		org	80h
kbd_start	dw	?		;start of keyboard buffer area
kbd_end		dw	?		;end of keyboard buffer area
rows_minus_1	db	?		;number of rows on screen minus one
char_height	dw	?		;charcter height in scan lines
info		db	?		;video mode options
info_3		db	?		;EGA feature bit switches
flags		db	?		;various flags
DCC_index	db	?		;Display Combination Code table addres
		org	0A8h
save_ptr	dd	?		;BIOS Video Save/Override Pointer
bda		ends			; Table addres

;Structure of Memory Control Block, also used for DOS Sub-Segments
;
;	Memory Block Type
;
;	"M"	normal Memory Block
;	"Z"	last Memory Block
;
;	Sub-Segment Type
;
;	"D"	Device Driver
;	"E"	Device Driver Appendage
;	"I"	IFS (Installable File System) Driver
;	"F"	FILES= control block storage area (for FILES > 5)
;	"X"	FCBS= control block storage area
;	"C"	BUFFERS /X Option (EMS workspace area)
;	"B"	BUFFERS= storage area
;	"L"	LASTDRIVE= current directory structure array
;	"S"	STACKS= code/data area, if present
;	"T"	INSTALL= initial install workspace (used in CONFIG.SYS)

mcb		segment at 0		;Memory Control Block or Arena Header
blk_flag	db	?		;flag 'M' or 'Z'
blk_owner	dw	?		;PSP of owner
blk_size	dw	?		;number of paragraphs owned
		org	8		;3 bytes are reserved
pgm_name	db	8 dup (?)	;name of owner witout extension
mcb		ends			;Devices have a similar control block

psp		segment	at 0		;Program Segment Prefix
int_20		dw	?		;code for Int 20h (CD 20)
seg_end_all_bl	dw	?		;segment, end of allocated block
		db	?		;reserved
call_f_disp	db	?		;far call to MS-DOS function
f_disp_addr	dd	?		; dispatcher (9A ?? ?? ?? ??) ?
term_addr	dd	?		;terminate handler (Int 22) address
prev_int_23	dd	?		;previous contents of Int 23 vector
prev_int_24	dd	?		;previous contents of Int 24 vector
parents_psp	dw	?		;PSP of parent process
handle_table	db	20 dup (?)	;file handle array
environment	dw	?		;segment address, environment
stack_ptr_stor	dd	?		;SS:SP on entry to last Int 21 function
hndl_array_size	dw	?		;handle array size
hndl_array_ptr	dd	?		;handle array pointer
previous_psp	dw	?		;pointer to previous PSP
					; (default FFFF:FFFF) ?
		db	20 dup (?)	;reserved
dos_f_disp	db	3 dup(?)	;DOS function dispatcher, Int 21 retf
					; (CD 21 CB)
		db	9 dup (?)	;reserved, can be used for XFCB #1
fcb_1		db	24h dup (?)	;default File Control Block #1
		org	6Ch
fcb_2		db	24h dup (?)	;default File Control Block #2
		org	80h
dta		label	byte		;default Disk Transfer Area (128 bytes)
command_tail	pascal_string <>	;command tail
psp		ends

;	The following macro builds strings using Turbo Pascal conventions.
;	That is, the string is stored with a preceding 'length byte'.
;
;	Input	Id  - Label to associate with the string
;		Msg - Contents of the string

TP_Str		macro	id, msg
local		endstr

ifb <id>
	display "Must pass name of string to TP_Str"
	err
else
	ifb <msg>
		id	db	256 dup (?)	;if message empty only space
	else					; is generated
		id	db	endstr - id - 1, &msg&
		endstr	label	byte
	endif
endif
		endm
