{$R-,S-,I-,D-,T-,V-,B-,N-,L- }

unit reloc;

{This unit relocates your program by loading a copy of it and releasing the
 first copy. Thus memory comes available under the program, enabling you to
 install interrupt handlers as low in memory as possible. These interrupt
 handlers then don't need a Segment Prefix (as long as they are not doing any
 disk IO) and will use as little emory as possible. Thus you can write
 interrupt handlers that take only 100 bytes.}

interface

var ver:                word;
    old_exitproc:       pointer;

procedure make_mem_block_independent (p: pointer);

implementation

uses psp_mcb, dos, env, memory;

var our_psp_ptr,
    parents_psp_ptr: ^psp_rec;
    term_addr:       ptrrec;
    old_stack_ofs,
    old_stack_seg:   word;

function compare_mcb_names(mcb1, mcb2: mcb): boolean;
var t: byte;
begin
  for t := 1 to 8 do
  begin
    if mcb1.ownername[t] <> mcb2.ownername[t]
    then
    begin
      compare_mcb_names:= false;
      exit
    end
    else if mcb1.ownername[t] = #0
    then
    begin
      compare_mcb_names:= true;
      exit
    end
  end;
  compare_mcb_names:= true
end;

procedure make_mem_block_independent (p: pointer);
var mem_block_mcb_ptr: ^mcb;
begin
  mem_block_mcb_ptr:= ptr(pred(ptrrec(p).psegment), 0);
  mem_block_mcb_ptr^.owner:= ptrrec(p).psegment;
  mem_block_mcb_ptr^.ownername:= mcb(ptr(pred(prefixseg), 0)^).ownername;
end;

procedure relocate_program;

procedure kill_parent; assembler;
asm
  mov   bx, [ptrrec(parents_psp_ptr).psegment]
  mov   ah, 50h              {set process id}
  int   21h
  push  bp
  mov   [old_stack_ofs], sp
  mov   [old_stack_seg], ss
  mov   ax, 4C00h
  int   21h
end;

procedure re_entry_point; assembler;
asm
  cli
  mov   ax, seg(@data)
  mov   ds, ax
  mov   ss, [old_stack_seg]
  mov   sp, [old_stack_ofs]
  sti
  pop   bp
end;

procedure my_terminator; assembler;
asm
  push  ax
  push  ax
  push  bp
  mov   bp, sp
  push  ax
  push  ds
  push  es
  pushf
  mov   ax, seg(@data)
  mov   ds, ax
  mov   es, [ptrrec(our_psp_ptr).psegment]
  mov   ah, 49h
  int   21h
  mov   ax, [term_addr.poffset]
  mov   [bp+2], ax
  mov   ax, [term_addr.psegment]
  mov   [bp+4], ax
  popf
  pop   es
  pop   ds
  pop   ax
  pop   bp
  retf
end;

procedure fake_residency; far;
var t: byte;
begin
  term_addr:= our_psp_ptr^.term_addr;
  pointer(our_psp_ptr^.term_addr):= @my_terminator;
  swapvectors;
  if exitcode <> 0 then halt(exitcode);
  for t:= 5 to pred(our_psp_ptr^.hndl_array_size) do {check for redirection}
  if byte(ptr(our_psp_ptr^.hndl_array_ptr.psegment,
              our_psp_ptr^.hndl_array_ptr.poffset + t)^) <> $FF
  then halt(exitcode);
  keep(exitcode)
end;

var environment:        word;
    tp1,
    tp2:                pointer;

begin
  swapvectors;
  our_psp_ptr:= ptr(prefixseg, 0);
  parents_psp_ptr:= ptr(our_psp_ptr^.parents_psp, 0);
  environment:= our_psp_ptr^.environment;
  if not compare_mcb_names(mcb(ptr(pred(prefixseg), 0)^),
                           mcb(ptr(pred(our_psp_ptr^.parents_psp), 0)^))
  then {we're original}
  begin
    setmemtop(heapptr);
    exec(getprogramname, our_psp_ptr^.command_tail);
    if doserror <> 0
    then
    begin
      if doserror = 8
      then writeln('Not enough memory for this Program.')
      else writeln('Undefined error reloading this Program.');
      halt($FF)
    end;
  end
  else {we're copy}
  begin
    getintvec($23, tp1);
    getintvec($24, tp2);
    our_psp_ptr^.prev_int_23:= parents_psp_ptr^.prev_int_23;
    our_psp_ptr^.prev_int_24:= parents_psp_ptr^.prev_int_24;
    setintvec($23, pointer(our_psp_ptr^.prev_int_23));
    setintvec($24, pointer(our_psp_ptr^.prev_int_24));
    pointer(parents_psp_ptr^.prev_int_23):= tp1;
    pointer(parents_psp_ptr^.prev_int_24):= tp2;
    our_psp_ptr^.term_addr:= parents_psp_ptr^.term_addr;
    our_psp_ptr^.parents_psp:= parents_psp_ptr^.parents_psp;
    parents_psp_ptr^.parents_psp:= ptrrec(our_psp_ptr).psegment;
    pointer(parents_psp_ptr^.term_addr):= @re_entry_point;
    kill_parent;
    asm
      mov   es, [environment]
      mov   ah, 49h
      int   21h
    end;
    swapvectors;
  end;
  old_exitproc:= exitproc;
  exitproc:= @fake_residency;
end;

begin
  ver:= dosversion;
  if Lo(Ver) < 3
  then
  begin
    writeln('This Program needs DOS version 3 or better.');
    halt($FF)
  end;
  relocate_program;
end.